/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.recording;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import org.audiochain.model.AudioProject;
import org.audiochain.model.AudioProjectAdapter;
import org.audiochain.model.AudioProjectLifecycleAdapter;
import org.audiochain.model.AudioTrack;
import org.audiochain.model.SourceAudioDevice;
import org.audiochain.ui.gui.BackgroundImageComponent;
import org.audiochain.ui.gui.CommonComponentFactory;
import org.audiochain.ui.gui.Dialog;
import org.audiochain.ui.gui.GuiSettings;

public class FileSequenceDialog extends JDialog {
	private static final long serialVersionUID = 1L;
	private static Map<AudioProject, FileSequenceDialog> projectToDialogInstancesMap;
	private static final int symbolPadding = 5;
	private final AudioProject project;
	private FileSequencePanel panel;
	private RecordingSettings settings;
	private GuiSettings guiSettings;
	private JScrollPane scroll;
	private Image trashImage;
	private Image triangleUpImage;
	private Image triangleDownImage;
	private int imageHeightHalf;
	private int imageWidth;

	private FileSequenceDialog(AudioProject project) {
		this.project = project;
		setTitle("File sequences");
		settings = RecordingSettings.getInstance();
		guiSettings = GuiSettings.getInstance();
		initGui();
		registerListeners();
	}

	private void initGui() {
		int x = settings.getFileSequenceDialogLocationX();
		int y = settings.getFileSequenceDialogLocationY();
		int width = settings.getFileSequenceDialogWidth();
		int height = settings.getFileSequenceDialogHeight();
		setBounds(x, y, width, height);
		panel = new FileSequencePanel();
		panel.setOpaque(false);
		JComponent contentPane = CommonComponentFactory.createDialogBackgroundComponent();
		contentPane.setLayout(new BorderLayout());
		scroll = CommonComponentFactory.createScrollPane(panel);
		contentPane.add(scroll, BorderLayout.CENTER);
		setContentPane(contentPane);
		trashImage = BackgroundImageComponent.readIcon("org/audiochain/devices/recording/trash-item.png").getImage();
		// assumed that all images have the same size
		imageHeightHalf = trashImage.getHeight(null) / 2;
		imageWidth = trashImage.getWidth(null);
		triangleUpImage = BackgroundImageComponent.readIcon("org/audiochain/ui/gui/triangleup-item.png").getImage();
		triangleDownImage = BackgroundImageComponent.readIcon("org/audiochain/ui/gui/triangledown-item.png").getImage();
		update();
	}

	private void update() {
		int preferredHeight = getAudioFileCount() * guiSettings.getFontSize() * 2 - scroll.getHorizontalScrollBar().getHeight();
		panel.setPreferredSize(new Dimension(450, preferredHeight));
		panel.repaint();
	}

	class FileSequencePanel extends JPanel {
		private static final long serialVersionUID = 1L;

		@Override
		public void paint(Graphics g) {
			final int panelWidth = getWidth();
			final int heightPerFile = getHeightPerFile();
			final int dyRect = heightPerFile / 4;
			final int dyRectHalf = dyRect / 2;
			final int rectHeight = heightPerFile - dyRect;
			final int fontSize = guiSettings.getFontSize();
			final int heightPerFileHalf = heightPerFile / 2;
			final int dyText = heightPerFileHalf + fontSize / 3;
			int y = 0;
			final long maxFrame = project.getMaxFrame();

			Collection<AudioTrack> audioTracks = project.getEnabledAudioTracks();
			for (AudioTrack audioTrack : audioTracks) {
				SourceAudioDevice sourceAudioDevice = audioTrack.getSourceAudioDevice();
				if (sourceAudioDevice instanceof RecordingAudioDevice) {
					RecordingAudioDevice device = (RecordingAudioDevice) sourceAudioDevice;
					AudioFileSequence activeAudioFileSequence = device.getActiveAudioFileSequence();
					if (activeAudioFileSequence != null) {
						Collection<AudioFile> audioFiles = activeAudioFileSequence.getAudioFiles();
						int fileCount = audioFiles.size();
						int fileNumber = 0;
						for (AudioFile audioFile : audioFiles) {
							fileNumber++;
							long frameOffset = audioFile.getFrameOffset();
							int x = (int) (frameOffset * panelWidth / maxFrame);
							long frameLength = audioFile.getFrameLength();
							int width = (int) (frameLength * panelWidth / maxFrame);
							g.setColor(guiSettings.getLightBackgroundColor());
							g.fillRect(x, y + dyRectHalf, width, rectHeight);
							g.setColor(guiSettings.getForegroundColor());
							g.drawString("Track " + audioTrack.getNumber() + " - " + audioFile.toString(), symbolPadding, y + dyText);
							int imageY = y + heightPerFileHalf - imageHeightHalf;
							g.drawImage(trashImage, panelWidth - 1 * imageWidth - symbolPadding, imageY, null);
							if (fileNumber < fileCount) {
								g.drawImage(triangleDownImage, panelWidth - 2 * imageWidth - symbolPadding, imageY, null);
							}
							if (fileNumber > 1) {
								g.drawImage(triangleUpImage, panelWidth - 3 * imageWidth - symbolPadding, imageY, null);
							}
							y += heightPerFile;
						}
					}
				}
			}

			long framePosition = project.getFramePosition();
			int rulerX = (int) (framePosition * panelWidth / maxFrame);
			g.drawLine(rulerX, 0, rulerX, getHeight());
		}
	}

	private int getAudioFileCount() {
		int audioFileCount = 0;
		Collection<AudioTrack> audioTracks = project.getEnabledAudioTracks();
		for (AudioTrack audioTrack : audioTracks) {
			SourceAudioDevice sourceAudioDevice = audioTrack.getSourceAudioDevice();
			if (sourceAudioDevice instanceof RecordingAudioDevice) {
				RecordingAudioDevice device = (RecordingAudioDevice) sourceAudioDevice;
				AudioFileSequence activeAudioFileSequence = device.getActiveAudioFileSequence();
				if (activeAudioFileSequence != null) {
					Collection<AudioFile> audioFiles = activeAudioFileSequence.getAudioFiles();
					audioFileCount += audioFiles.size();
				}
			}
		}
		return audioFileCount;
	}

	private int getHeightPerFile() {
		int panelHeight = panel.getHeight();
		int audioFileCount = getAudioFileCount();
		if (audioFileCount > 0) {
			return panelHeight / audioFileCount;
		}
		return panelHeight;
	}

	private void registerListeners() {
		RecordingAudioDeviceListener deviceListener = new RecordingAudioDeviceAdapter() {
			@Override
			public void audioFileSequenceActivated(AudioFileSequence sequence) {
				update();
			}

			@Override
			public void audioFileSequenceAdded(AudioFileSequence sequence) {
				update();
			}

			@Override
			public void audioFileSequenceRemoved(AudioFileSequence sequence) {
				update();
			}

			@Override
			public void audioFileSequenceUpdated(AudioFileSequence sequence) {
				update();
			}

			@Override
			public void audioFileFrameOffsetChanged(AudioFile audioFile, long oldFrameOffset, long newFrameOffset) {
				update();
			}

			@Override
			public void newAudioFileAdded(AudioFile audioFile) {
				update();
			}

			@Override
			public void audioFileDeleted(AudioFile audioFile) {
				update();
			}

			@Override
			public void audioFileSequenceDeleted(AudioFileSequence sequence) {
				update();
			}
		};
		Collection<AudioTrack> audioTracks = project.getAudioTracks();
		for (AudioTrack audioTrack : audioTracks) {
			SourceAudioDevice sourceAudioDevice = audioTrack.getSourceAudioDevice();
			if (sourceAudioDevice instanceof RecordingAudioDevice) {
				RecordingAudioDevice device = (RecordingAudioDevice) sourceAudioDevice;
				device.addRecordingAudioDeviceListener(deviceListener);
			}
		}
		project.addAudioProjectChangeListener(new AudioProjectAdapter() {

			@Override
			public void audioTrackAdded(AudioTrack track) {
				update();
			}

			@Override
			public void audioTrackRemoved(AudioTrack track) {
				update();
			}

			@Override
			public void audioProjectFrameLengthChanged() {
				update();
			}

			@Override
			public void audioProjectFramePositionChanged(long oldValue, long newValue) throws IOException {
				update();
			}

			@Override
			public void audioTrackMoved(AudioTrack track) {
				update();
			}

		});
		project.addAudioProjectLifecycleListener(new AudioProjectLifecycleAdapter() {
			@Override
			public void audioProjectClosed(AudioProject audioProject) {
				getFileSequenceDialogInstances().remove(audioProject);
				storeSettings();
				dispose();
			}
		});
		addComponentListener(new ComponentAdapter() {
			@Override
			public void componentHidden(ComponentEvent e) {
				getFileSequenceDialogInstances().remove(project);
				storeSettings();
			}
		});
		panel.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				int panelWidth = panel.getWidth();
				int clickedY = e.getY();
				int heightPerFile = getHeightPerFile();
				final int heightPerFileHalf = heightPerFile / 2;
				int clickedIndex = clickedY / heightPerFile;
				int index = 0;
				int trashX = panelWidth - 1 * imageWidth - symbolPadding;
				int triangleDownX = panelWidth - 2 * imageWidth - symbolPadding;
				int triangleUpX = panelWidth - 3 * imageWidth - symbolPadding;

				Collection<AudioTrack> audioTracks = project.getEnabledAudioTracks();
				for (AudioTrack audioTrack : audioTracks) {
					SourceAudioDevice sourceAudioDevice = audioTrack.getSourceAudioDevice();
					if (sourceAudioDevice instanceof RecordingAudioDevice) {
						RecordingAudioDevice device = (RecordingAudioDevice) sourceAudioDevice;
						AudioFileSequence sequence = device.getActiveAudioFileSequence();
						if (sequence != null) {
							Collection<AudioFile> audioFiles = sequence.getAudioFiles();
							int fileCount = audioFiles.size();
							int fileNumber = 0;
							for (AudioFile audioFile : audioFiles) {
								fileNumber++;
								if (index == clickedIndex) {
									int imageY = index * heightPerFile + heightPerFileHalf - imageHeightHalf;
									if (withinImageBounds(e, trashX, imageY)) {
										if (Dialog.okCancel(FileSequenceDialog.this, "Delete Audio File", "The Audio File '" + audioFile + "' will be deleted.")) {
											device.deleteAudioFile(audioFile);
										}
									} else if (fileNumber < fileCount && withinImageBounds(e, triangleDownX, imageY)) {
										device.incrementAudioFilePosition(sequence, audioFile);
									} else if (fileNumber > 1 && withinImageBounds(e, triangleUpX, imageY)) {
										device.decrementAudioFilePosition(sequence, audioFile);
									}
								}
								index++;
							}
						}
					}
				}
			}

			private boolean withinImageBounds(MouseEvent e, int imageX, int imageY) {
				int x = e.getX();
				int y = e.getY();
				return x >= imageX && x < imageX + imageWidth && y >= imageY && y < imageY + 2 * imageHeightHalf;
			}
		});
	}

	private void storeSettings() {
		settings.setFileSequenceDialogLocationX(FileSequenceDialog.this.getX());
		settings.setFileSequenceDialogLocationY(FileSequenceDialog.this.getY());
		settings.setFileSequenceDialogWidth(FileSequenceDialog.this.getWidth());
		settings.setFileSequenceDialogHeight(FileSequenceDialog.this.getHeight());
		settings.store();
	}

	private static Map<AudioProject, FileSequenceDialog> getFileSequenceDialogInstances() {
		if (projectToDialogInstancesMap == null) {
			projectToDialogInstancesMap = new HashMap<AudioProject, FileSequenceDialog>();
		}
		return projectToDialogInstancesMap;
	}

	public static FileSequenceDialog getInstance(AudioProject project) {
		FileSequenceDialog ret = getFileSequenceDialogInstances().get(project);
		if (ret == null) {
			ret = new FileSequenceDialog(project);
			getFileSequenceDialogInstances().put(project, ret);
		}
		return ret;
	}
}
